import hashlib
from collections import defaultdict
from collections.abc import Callable, Hashable
from dataclasses import dataclass
from datetime import datetime, timezone
from functools import partial
from pathlib import Path
from typing import Any, cast

import rdflib

from cognite.neat._client import NeatClient
from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._issues import IssueList, catch_issues
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules._shared import ReadRules, Rules, T_VerifiedRules, VerifiedRules
from cognite.neat._rules.exporters import BaseExporter
from cognite.neat._rules.exporters._base import CDFExporter, T_Export
from cognite.neat._rules.importers import BaseImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.models._base_input import InputRules
from cognite.neat._rules.transformers import RulesTransformer
from cognite.neat._utils.upload import UploadResultList

from ._provenance import EMPTY_ENTITY, UNKNOWN_AGENT, Activity, Agent, Change, Entity, Provenance
from .exceptions import EmptyStore, InvalidInputOperation


@dataclass(frozen=True)
class ModelEntity(Entity):
    result: Rules | None = None

    @property
    def display_name(self) -> str:
        if self.result is None:
            return "Failed"
        if isinstance(self.result, ReadRules):
            if self.result.rules is None:
                return "FailedRead"
            return self.result.rules.display_type_name()
        else:
            return self.result.display_type_name()


@dataclass(frozen=True)
class OutcomeEntity(Entity):
    result: UploadResultList | Path | str | None = None


class NeatRulesStore:
    def __init__(self) -> None:
        self.provenance = Provenance()
        self.exports_by_source_entity_id: dict[rdflib.URIRef, list[Change]] = defaultdict(list)
        self.pruned_by_source_entity_id: dict[rdflib.URIRef, list[Provenance]] = defaultdict(list)
        self._last_outcome: UploadResultList | None = None
        self._iteration_by_id: dict[Hashable, int] = {}

    def calculate_provenance_hash(self, shorten: bool = True) -> str:
        sha256_hash = hashlib.sha256()
        for change in self.provenance:
            for id_ in [change.agent.id_, change.activity.id_, change.target_entity.id_, change.source_entity.id_]:
                sha256_hash.update(str(id_).encode("utf-8"))
        calculated_hash = sha256_hash.hexdigest()
        if shorten:
            return calculated_hash[:8]
        return calculated_hash

    def import_(self, importer: BaseImporter) -> IssueList:
        agent = importer.agent
        source_entity = Entity(
            was_attributed_to=UNKNOWN_AGENT,
            id_=importer.source_uri,
        )
        return self._do_activity(importer.to_rules, agent, source_entity, importer.description)[1]

    def transform(self, *transformer: RulesTransformer) -> IssueList:
        if not self.provenance:
            raise EmptyStore()

        all_issues = IssueList()
        for item in transformer:
            last_change = self.provenance[-1]
            source_entity = last_change.target_entity
            if not isinstance(source_entity, ModelEntity):
                # Todo: Provenance should be of an entity type
                raise ValueError("Bug in neat: The last entity in the provenance is not a model entity.")
            transformer_input = source_entity.result

            if not item.is_valid_input(transformer_input):
                raise InvalidInputOperation(expected=item.transform_type_hint(), got=type(transformer_input))

            transform_issues = self._do_activity(
                partial(item.transform, rules=transformer_input),
                item.agent,
                source_entity,
                item.description,
            )[1]
            all_issues.extend(transform_issues)
        return all_issues

    def export(self, exporter: BaseExporter[T_VerifiedRules, T_Export]) -> T_Export:
        last_change = self.provenance[-1]
        source_entity = last_change.target_entity
        if not isinstance(source_entity, ModelEntity):
            # Todo: Provenance should be of an entity type
            raise ValueError("Bug in neat: The last entity in the provenance is not a model entity.")
        expected_types = exporter.source_types()
        if not any(isinstance(source_entity.result, type_) for type_ in expected_types):
            raise InvalidInputOperation(expected=expected_types, got=type(source_entity.result))

        agent = exporter.agent
        start = datetime.now(timezone.utc)
        with catch_issues() as issue_list:
            # Validate the type of the result
            result = exporter.export(source_entity.result)  # type: ignore[arg-type]
        end = datetime.now(timezone.utc)
        target_id = DEFAULT_NAMESPACE["export-result"]
        activity = Activity(
            was_associated_with=agent,
            ended_at_time=end,
            started_at_time=start,
            used=source_entity,
        )
        target_entity = OutcomeEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            result=type(result).__name__,
            issues=issue_list,
            id_=target_id,
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=exporter.description,
            source_entity=source_entity,
        )
        self.exports_by_source_entity_id[source_entity.id_].append(change)
        return result

    def export_to_file(self, exporter: BaseExporter, path: Path) -> None:
        last_change = self.provenance[-1]
        source_entity = last_change.target_entity
        if not isinstance(source_entity, ModelEntity):
            # Todo: Provenance should be of an entity type
            raise ValueError("Bug in neat: The last entity in the provenance is not a model entity.")
        expected_types = exporter.source_types()
        if not any(isinstance(source_entity.result, type_) for type_ in expected_types):
            raise InvalidInputOperation(expected=expected_types, got=type(source_entity.result))
        target_id = DEFAULT_NAMESPACE[path.name]
        agent = exporter.agent
        start = datetime.now(timezone.utc)
        with catch_issues() as issue_list:
            exporter.export_to_file(source_entity.result, path)
        end = datetime.now(timezone.utc)

        activity = Activity(
            was_associated_with=agent,
            ended_at_time=end,
            started_at_time=start,
            used=source_entity,
        )
        target_entity = OutcomeEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            result=path,
            issues=issue_list,
            id_=target_id,
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=exporter.description,
            source_entity=source_entity,
        )
        self.exports_by_source_entity_id[source_entity.id_].append(change)

    def export_to_cdf(self, exporter: CDFExporter, client: NeatClient, dry_run: bool) -> UploadResultList:
        last_change = self.provenance[-1]
        source_entity = last_change.target_entity
        if not isinstance(source_entity, ModelEntity):
            # Todo: Provenance should be of an entity type
            raise ValueError("Bug in neat: The last entity in the provenance is not a model entity.")
        expected_types = exporter.source_types()
        if not any(isinstance(source_entity.result, type_) for type_ in expected_types):
            raise InvalidInputOperation(expected=expected_types, got=type(source_entity.result))

        agent = exporter.agent
        start = datetime.now(timezone.utc)
        target_id = DEFAULT_NAMESPACE["upload-result"]
        result: UploadResultList | None = None
        with catch_issues() as issue_list:
            result = exporter.export_to_cdf(source_entity.result, client, dry_run)
        end = datetime.now(timezone.utc)

        activity = Activity(
            was_associated_with=agent,
            ended_at_time=end,
            started_at_time=start,
            used=source_entity,
        )
        target_entity = OutcomeEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            result=result,
            issues=issue_list,
            id_=target_id,
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=exporter.description,
            source_entity=source_entity,
        )
        self.exports_by_source_entity_id[source_entity.id_].append(change)
        self._last_outcome = result
        return result

    def prune_until_compatible(self, transformer: RulesTransformer) -> list[Change]:
        """Prune the provenance until the last successful entity is compatible with the transformer.

        Args:
            transformer: The transformer to check compatibility with.

        Returns:
            The changes that were pruned.
        """
        pruned_candidates: list[Change] = []
        for change in reversed(self.provenance):
            if not isinstance(change.target_entity, ModelEntity):
                continue
            if not transformer.is_valid_input(change.target_entity.result):
                pruned_candidates.append(change)
            else:
                break
        else:
            raise NeatValueError("No compatible entity found in the provenance.")
        if not pruned_candidates:
            return []
        self.provenance = self.provenance[: -len(pruned_candidates)]
        pruned_candidates.reverse()
        self.pruned_by_source_entity_id[self.provenance[-1].target_entity.id_].append(Provenance(pruned_candidates))
        return pruned_candidates

    def _export(self, action: Callable[[Any], Any], agent: Agent, description: str) -> Any:
        last_entity: ModelEntity | None = None
        for change in reversed(self.provenance):
            if isinstance(change.target_entity, ModelEntity) and isinstance(change.target_entity.result, DMSRules):
                last_entity = change.target_entity
                break
        if last_entity is None:
            raise NeatValueError("No verified DMS rules found in the provenance.")
        rules = last_entity.result
        result, _ = self._do_activity(lambda: action(rules), agent, last_entity, description)
        return result

    def _do_activity(
        self, action: Callable[[], Rules | None], agent: Agent, source_entity: Entity, description: str
    ) -> tuple[Any, IssueList]:
        start = datetime.now(timezone.utc)
        result: Rules | None = None
        with catch_issues() as issue_list:
            result = action()
        end = datetime.now(timezone.utc)

        activity = Activity(
            was_associated_with=agent,
            ended_at_time=end,
            started_at_time=start,
            used=source_entity,
        )
        target_entity = ModelEntity(
            was_attributed_to=agent,
            was_generated_by=activity,
            result=result,
            issues=issue_list,
            id_=self._create_id(result),
        )
        change = Change(
            agent=agent,
            activity=activity,
            target_entity=target_entity,
            description=description,
            source_entity=source_entity,
        )

        self.provenance.append(change)
        return result, issue_list

    def _create_id(self, result: Any) -> rdflib.URIRef:
        identifier: rdflib.URIRef
        if result is None:
            identifier = EMPTY_ENTITY.id_
        elif isinstance(result, ReadRules):
            if result.rules is None:
                identifier = EMPTY_ENTITY.id_
            else:
                identifier = result.rules.metadata.identifier
        elif isinstance(result, VerifiedRules):
            identifier = result.metadata.identifier
        else:
            identifier = DEFAULT_NAMESPACE["unknown-entity"]

        if identifier not in self._iteration_by_id:
            self._iteration_by_id[identifier] = 1
            return identifier
        self._iteration_by_id[identifier] += 1
        return identifier + f"/Iteration_{self._iteration_by_id[identifier]}"

    def get_last_entity(self) -> ModelEntity:
        if not self.provenance:
            raise NeatValueError("No entity found in the provenance.")
        return cast(ModelEntity, self.provenance[-1].target_entity)

    def get_last_successful_entity(self) -> ModelEntity:
        for change in reversed(self.provenance):
            if isinstance(change.target_entity, ModelEntity) and change.target_entity.result:
                return change.target_entity
        raise NeatValueError("No successful entity found in the provenance.")

    @property
    def has_unverified_rules(self) -> bool:
        return any(
            isinstance(change.target_entity, ModelEntity)
            and isinstance(change.target_entity.result, ReadRules)
            and change.target_entity.result.rules is not None
            for change in self.provenance
        )

    @property
    def has_verified_rules(self) -> bool:
        return any(
            isinstance(change.target_entity, ModelEntity)
            and isinstance(change.target_entity.result, DMSRules | InformationRules)
            for change in self.provenance
        )

    @property
    def last_unverified_rule(self) -> InputRules:
        for change in reversed(self.provenance):
            if (
                isinstance(change.target_entity, ModelEntity)
                and isinstance(change.target_entity.result, ReadRules)
                and change.target_entity.result.rules is not None
            ):
                return change.target_entity.result.rules

        raise NeatValueError("No unverified rule found in the provenance.")

    @property
    def last_verified_rule(self) -> DMSRules | InformationRules:
        for change in reversed(self.provenance):
            if isinstance(change.target_entity, ModelEntity) and isinstance(
                change.target_entity.result, DMSRules | InformationRules
            ):
                return change.target_entity.result
        raise NeatValueError("No verified rule found in the provenance.")

    @property
    def last_verified_dms_rules(self) -> DMSRules:
        for change in reversed(self.provenance):
            if isinstance(change.target_entity, ModelEntity) and isinstance(change.target_entity.result, DMSRules):
                return change.target_entity.result
        raise NeatValueError("No verified DMS rules found in the provenance.")

    @property
    def last_issues(self) -> IssueList:
        if not self.provenance:
            raise NeatValueError("No issues found in the provenance.")
        return self.provenance[-1].target_entity.issues

    @property
    def last_outcome(self) -> UploadResultList:
        if self._last_outcome is not None:
            return self._last_outcome
        raise NeatValueError("No outcome found in the provenance.")



import warnings
from abc import ABC, abstractmethod
from collections.abc import Iterator
from contextlib import contextmanager, suppress
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, Literal

from pydantic import ValidationError
from rdflib import URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._issues import IssueList, NeatError, NeatWarning
from cognite.neat._rules._shared import ReadRules, T_InputRules
from cognite.neat._utils.auxiliary import class_html_doc

if TYPE_CHECKING:
    from cognite.neat._store._provenance import Agent as ProvenanceAgent


class BaseImporter(ABC, Generic[T_InputRules]):
    """
    BaseImporter class which all importers inherit from.
    """

    @abstractmethod
    def to_rules(self) -> ReadRules[T_InputRules]:
        """Creates `Rules` object from the data for target role."""
        raise NotImplementedError()

    def _default_metadata(self) -> dict[str, Any]:
        creator = "UNKNOWN"
        with suppress(KeyError, ImportError):
            import getpass

            creator = getpass.getuser()

        return {
            "schema": "partial",
            "space": "neat",
            "external_id": "NeatImportedDataModel",
            "version": "0.1.0",
            "name": "Neat Imported Data Model",
            "created": datetime.now().replace(microsecond=0).isoformat(),
            "updated": datetime.now().replace(microsecond=0).isoformat(),
            "creator": creator,
            "description": f"Imported using {type(self).__name__}",
        }

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)

    @property
    def agent(self) -> "ProvenanceAgent":
        """Provenance agent for the importer."""
        from cognite.neat._store._provenance import Agent as ProvenanceAgent

        return ProvenanceAgent(id_=DEFAULT_NAMESPACE[f"agent/{type(self).__name__}"])

    @property
    def description(self) -> str:
        return "MISSING DESCRIPTION"

    @property
    def source_uri(self) -> URIRef:
        return DEFAULT_NAMESPACE["UNKNOWN"]


class _FutureResult:
    def __init__(self) -> None:
        self._result: Literal["success", "failure", "pending"] = "pending"

    @property
    def result(self) -> Literal["success", "failure", "pending"]:
        return self._result


@contextmanager
def _handle_issues(
    issues: IssueList,
    error_cls: type[NeatError] = NeatError,
    warning_cls: type[NeatWarning] = NeatWarning,
    error_args: dict[str, Any] | None = None,
) -> Iterator[_FutureResult]:
    """This is an internal help function to handle issues and warnings.

    Args:
        issues: The issues list to append to.
        error_cls: The class used to convert errors to issues.
        warning_cls:  The class used to convert warnings to issues.

    Returns:
        FutureResult: A future result object that can be used to check the result of the context manager.
    """
    with warnings.catch_warnings(record=True) as warning_logger:
        warnings.simplefilter("always")
        future_result = _FutureResult()
        try:
            yield future_result
        except ValidationError as e:
            issues.extend(error_cls.from_errors(e.errors(), **(error_args or {})))  # type: ignore[arg-type]
            future_result._result = "failure"
        else:
            future_result._result = "success"
        finally:
            if warning_logger:
                issues.extend([warning_cls.from_warning(warning) for warning in warning_logger])  # type: ignore[misc]



from collections.abc import Iterable, Iterator
from typing import cast

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.instances import Instance, PropertyValue
from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat._constants import DEFAULT_SPACE_URI
from cognite.neat._issues.errors import ResourceRetrievalError
from cognite.neat._shared import Triple

from ._base import BaseExtractor


class DMSExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusion DMS instances into Neat.

    Args:
        items: The items to extract.
        total: The total number of items to extract. If provided, this will be used to estimate the progress.
        limit: The maximum number of items to extract.
        overwrite_namespace: If provided, this will overwrite the space of the extracted items.
    """

    def __init__(
        self,
        items: Iterable[Instance],
        total: int | None = None,
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
    ) -> None:
        self.items = items
        self.total = total
        self.limit = limit
        self.overwrite_namespace = overwrite_namespace

    @classmethod
    def from_data_model(
        cls, client: CogniteClient, data_model: DataModelIdentifier, limit: int | None = None
    ) -> "DMSExtractor":
        """Create an extractor from a data model.

        Args:
            client: The Cognite client to use.
            data_model: The data model to extract.
            limit: The maximum number of instances to extract.
        """
        retrieved = client.data_modeling.data_models.retrieve(data_model, inline_views=True)
        if not retrieved:
            raise ResourceRetrievalError(dm.DataModelId.load(data_model), "data model", "Data Model is missing in CDF")
        return cls.from_views(client, retrieved.latest_version().views, limit)

    @classmethod
    def from_views(cls, client: CogniteClient, views: Iterable[dm.View], limit: int | None = None) -> "DMSExtractor":
        """Create an extractor from a set of views.

        Args:
            client: The Cognite client to use.
            views: The views to extract.
            limit: The maximum number of instances to extract.
        """
        return cls(_InstanceIterator(client, views), total=None, limit=limit)

    def extract(self) -> Iterable[Triple]:
        for count, item in enumerate(self.items, 1):
            if self.limit and count > self.limit:
                break
            yield from self._extract_instance(item)

    def _extract_instance(self, instance: Instance) -> Iterable[Triple]:
        if isinstance(instance, dm.Edge):
            if not instance.properties:
                yield (
                    self._as_uri_ref(instance.start_node),
                    self._as_uri_ref(instance.type),
                    self._as_uri_ref(instance.end_node),
                )
                return
            else:
                # If the edge has properties, we create a node for the edge and connect it to the start and end nodes.
                id_ = self._as_uri_ref(instance)
                yield id_, RDF.type, self._as_uri_ref(instance.type)
                yield id_, RDF.type, self._get_namespace(instance.space).Edge
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "startNode")),
                    self._as_uri_ref(instance.start_node),
                )
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "endNode")),
                    self._as_uri_ref(instance.end_node),
                )

        elif isinstance(instance, dm.Node):
            id_ = self._as_uri_ref(instance)
            if instance.type:
                type_ = self._as_uri_ref(cast(dm.DirectRelationReference, instance.type))
            else:
                type_ = self._get_namespace(instance.space).Node

            yield id_, RDF.type, type_
        else:
            raise NotImplementedError(f"Unknown instance type {type(instance)}")

        for view_id, properties in instance.properties.items():
            namespace = self._get_namespace(view_id.space)
            for key, value in properties.items():
                for object_ in self._get_objects(value):
                    yield id_, namespace[key], object_

    def _get_objects(self, value: PropertyValue) -> Iterable[Literal | URIRef]:
        if isinstance(value, str | float | bool | int):
            yield Literal(value)
        elif isinstance(value, dict) and "space" in value and "externalId" in value:
            yield self._as_uri_ref(dm.DirectRelationReference.load(value))
        elif isinstance(value, dict):
            # This object is a json object.
            yield Literal(str(value), datatype=XSD._NS["json"])
        elif isinstance(value, list):
            for item in value:
                yield from self._get_objects(item)

    def _as_uri_ref(self, instance: Instance | dm.DirectRelationReference) -> URIRef:
        return self._get_namespace(instance.space)[instance.external_id]

    def _get_namespace(self, space: str) -> Namespace:
        if self.overwrite_namespace:
            return self.overwrite_namespace
        return Namespace(DEFAULT_SPACE_URI.format(space=space))


class _InstanceIterator(Iterator[Instance]):
    def __init__(self, client: CogniteClient, views: Iterable[dm.View]):
        self.client = client
        self.views = views

    def __iter__(self) -> Iterator[Instance]:
        return self

    def __next__(self) -> Instance:  # type: ignore[misc]
        for view in self.views:
            # All nodes and edges with properties
            yield from self.client.data_modeling.instances(chunk_size=None, instance_type="node", sources=[view])
            yield from self.client.data_modeling.instances(chunk_size=None, instance_type="edge", sources=[view])

            for prop in view.properties.values():
                if isinstance(prop, dm.EdgeConnection):
                    # Get all edges with properties
                    yield from self.client.data_modeling.instances(
                        chunk_size=None,
                        instance_type="edge",
                        filter=dm.filters.Equals(
                            ["edge", "type"], {"space": prop.type.space, "externalId": prop.type.external_id}
                        ),
                    )



from cognite.client import data_modeling as dm

from cognite.neat._constants import COGNITE_MODELS
from cognite.neat._issues import IssueList
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.transformers import SetIDDMSModel

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class SetAPI:
    """Used to change the name of the data model from a data model id defined by neat to a user specified name."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def data_model_id(self, new_model_id: dm.DataModelId | tuple[str, str, str]) -> IssueList:
        """Sets the data model ID of the latest verified data model. Set the data model id as a tuple of strings
        following the template (<data_model_space>, <data_model_name>, <data_model_version>).

        Example:
            Set a new data model id:
            ```python
            neat.set.data_model_id(("my_data_model_space", "My_Data_Model", "v1"))
            ```
        """
        rules = self._state.rule_store.get_last_successful_entity().result
        if isinstance(rules, DMSRules):
            if rules.metadata.as_data_model_id() in COGNITE_MODELS:
                raise NeatSessionError(
                    "Cannot change the data model ID of a Cognite Data Model in NeatSession"
                    " due to temporarily issue with the reverse direct relation interpretation"
                )
        return self._state.rule_transform(SetIDDMSModel(new_model_id))



from cognite.neat._client import NeatClient
from cognite.neat._issues import IssueList
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.models.mapping import load_classic_to_core_mapping
from cognite.neat._rules.transformers import AsParentPropertyId, IncludeReferenced, RuleMapper

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class MappingAPI:
    def __init__(self, state: SessionState, client: NeatClient | None = None):
        self.data_model = DataModelMappingAPI(state, client)


@session_class_wrapper
class DataModelMappingAPI:
    def __init__(self, state: SessionState, client: NeatClient | None = None):
        self._state = state
        self._client = client

    def classic_to_core(self, company_prefix: str, use_parent_property_name: bool = True) -> IssueList:
        """Map classic types to core types.

        Note this automatically creates an extended CogniteCore model.

        Args:
            company_prefix: Prefix used for all extended CogniteCore types.
            use_parent_property_name: Whether to use the parent property name in the extended CogniteCore model.
                See below for more information.

        If you extend CogniteAsset, with for example, ClassicAsset. You will map the property `parentId` to `parent`.
        If you set `user_parent_property_name` to True, the `parentId` will be renamed to `parent` after the
        mapping is done. If you set it to False, the property will remain `parentId`.

        Example:
            ```python
            neat.mapping.classic_to_core(company_prefix="WindFarmX", use_parent_property_name=True)
            ```
        """
        rules = self._state.rule_store.get_last_successful_entity().result
        if not isinstance(rules, DMSRules):
            # Todo better hint of what you should do.
            raise NeatSessionError(f"Expected DMSRules, got {type(rules)}")

        if self._client is None:
            raise NeatSessionError("Client is required to map classic to core")

        transformers = [
            RuleMapper(load_classic_to_core_mapping(company_prefix, rules.metadata.space, rules.metadata.version)),
            IncludeReferenced(self._client),
        ]
        if use_parent_property_name:
            transformers.append(AsParentPropertyId(self._client))
        return self._state.rule_transform(*transformers)



from dataclasses import dataclass, field
from typing import Literal, cast

from cognite.neat._issues import IssueList
from cognite.neat._rules.importers import BaseImporter, InferenceImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.transformers import RulesTransformer, ToExtensionModel
from cognite.neat._store import NeatGraphStore, NeatRulesStore
from cognite.neat._store._rules_store import ModelEntity
from cognite.neat._utils.rdf_ import uri_display_name
from cognite.neat._utils.text import humanize_collection
from cognite.neat._utils.upload import UploadResultList

from .exceptions import NeatSessionError


class SessionState:
    def __init__(self, store_type: Literal["memory", "oxigraph"]) -> None:
        self.instances = InstancesState(store_type)
        self.rule_store = NeatRulesStore()
        self.last_reference: DMSRules | InformationRules | None = None

    def rule_transform(self, *transformer: RulesTransformer) -> IssueList:
        if not transformer:
            raise NeatSessionError("No transformers provided.")
        first_transformer = transformer[0]
        pruned = self.rule_store.prune_until_compatible(first_transformer)
        if pruned:
            type_hint = first_transformer.transform_type_hint()
            action = uri_display_name(first_transformer.agent.id_)
            location = cast(ModelEntity, self.rule_store.provenance[-1].target_entity).display_name
            expected = humanize_collection([hint.display_type_name() for hint in type_hint])  # type: ignore[attr-defined]
            step_str = "step" if len(pruned) == 1 else "steps"
            print(
                f"The {action} actions expects a {expected}. "
                f"Moving back {len(pruned)} {step_str} to the last {location}."
            )
        if (
            any(isinstance(t, ToExtensionModel) for t in transformer)
            and isinstance(self.rule_store.provenance[-1].target_entity, ModelEntity)
            and isinstance(self.rule_store.provenance[-1].target_entity.result, DMSRules | InformationRules)
        ):
            self.last_reference = self.rule_store.provenance[-1].target_entity.result

        start = cast(ModelEntity, self.rule_store.provenance[-1].target_entity).display_name
        issues = self.rule_store.transform(*transformer)
        end = cast(ModelEntity, self.rule_store.provenance[-1].target_entity).display_name
        issues.action = f"{start} &#8594; {end}"
        issues.hint = "Use the .inspect.issues() for more details."
        return issues

    def rule_import(self, importer: BaseImporter) -> IssueList:
        issues = self.rule_store.import_(importer)
        result = cast(ModelEntity, self.rule_store.provenance[-1].target_entity).display_name
        if isinstance(importer, InferenceImporter):
            issues.action = f"Inferred {result}"
        else:
            issues.action = f"Read {result}"
        if issues:
            issues.hint = "Use the .inspect.issues() for more details."
        return issues


@dataclass
class InstancesState:
    store_type: Literal["memory", "oxigraph"]
    issue_lists: list[IssueList] = field(default_factory=list)
    outcome: list[UploadResultList] = field(default_factory=list)
    _store: NeatGraphStore | None = field(init=False, default=None)

    @property
    def store(self) -> NeatGraphStore:
        if not self.has_store:
            if self.store_type == "oxigraph":
                self._store = NeatGraphStore.from_oxi_store()
            else:
                self._store = NeatGraphStore.from_memory_store()
        return cast(NeatGraphStore, self._store)

    @property
    def has_store(self) -> bool:
        return self._store is not None

    @property
    def last_outcome(self) -> UploadResultList:
        if not self.outcome:
            raise NeatSessionError(
                "No outcome available. Try using [bold].to.cdf.instances[/bold] to upload a data minstances."
            )
        return self.outcome[-1]



from typing import cast
from urllib.parse import quote

from rdflib import Graph, Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._rules.analysis import InformationAnalysis
from cognite.neat._rules.models._rdfpath import RDFPath, SingleProperty
from cognite.neat._rules.models.information import InformationRules
from cognite.neat._shared import Triple
from cognite.neat._utils.rdf_ import get_namespace, remove_namespace_from_uri

from ._base import BaseTransformer, BaseTransformerStandardised, RowTransformationOutput


class ReduceHopTraversal(BaseTransformer):
    """ReduceHopTraversal is a transformer that reduces the number of hops to direct connection."""

    ...


# TODO: Standardise
class AddSelfReferenceProperty(BaseTransformer):
    description: str = "Adds property that contains id of reference to all references of given class in Rules"
    _use_only_once: bool = True
    _need_changes = frozenset({})
    _ref_template: str = """SELECT ?s WHERE {{?s a <{type_}>}}"""

    def __init__(
        self,
        rules: InformationRules,
    ):
        self.rules = rules
        self.properties = InformationAnalysis(rules).all_reference_transformations()

    def transform(self, graph: Graph) -> None:
        for property_ in self.properties:
            prefix = property_.transformation.traversal.class_.prefix
            suffix = property_.transformation.traversal.class_.suffix

            namespace = self.rules.prefixes[prefix] if prefix in self.rules.prefixes else self.rules.metadata.namespace

            for (reference,) in graph.query(self._ref_template.format(type_=namespace[suffix])):  # type: ignore [misc]
                graph.add(
                    (
                        reference,
                        self.rules.metadata.namespace[property_.property_],
                        reference,
                    )
                )

            traversal = SingleProperty.from_string(
                class_=property_.view.id,
                property_=f"{self.rules.metadata.prefix}:{property_.property_}",
            )

            property_.transformation = RDFPath(traversal=traversal)


class MakeConnectionOnExactMatch(BaseTransformerStandardised):
    description: str = "Adds property that contains id of reference to all references of given class in Rules"
    _use_only_once: bool = True
    _need_changes = frozenset({})

    def __init__(
        self,
        subject_type: URIRef,
        subject_predicate: URIRef,
        object_type: URIRef,
        object_predicate: URIRef,
        connection: URIRef | str | None = None,
        limit: int | None = None,
    ):
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.object_type = object_type
        self.object_predicate = object_predicate
        subject_namespace = Namespace(get_namespace(subject_type))
        self.connection = (
            subject_namespace[quote(connection.strip())]
            if isinstance(connection, str)
            else connection or subject_namespace[remove_namespace_from_uri(self.object_type).lower()]
        )

        self.limit = limit

    def _iterate_query(self) -> str:
        query = """SELECT DISTINCT ?subject ?object
                            WHERE {{
                                ?subject a <{subject_type}> ;
                                        <{subject_predicate}> ?value .
                                ?object a <{object_type}> ;
                                        <{object_predicate}> ?value .
                            }}"""

        if self.limit and isinstance(self.limit, int) and self.limit > 0:
            query += f" LIMIT {self.limit}"

        return query.format(
            subject_type=self.subject_type,
            subject_predicate=self.subject_predicate,
            object_type=self.object_type,
            object_predicate=self.object_predicate,
        )

    def _count_query(self) -> str:
        query = """SELECT (COUNT(DISTINCT ?subject) as ?count)
                    WHERE {{
                        ?subject a <{subject_type}> ;
                                <{subject_predicate}> ?value .
                        ?object a <{object_type}> ;
                                <{object_predicate}> ?value .
                    }}"""

        if self.limit and isinstance(self.limit, int) and self.limit > 0:
            query += f" LIMIT {self.limit}"

        return query.format(
            subject_type=self.subject_type,
            subject_predicate=self.subject_predicate,
            object_type=self.object_type,
            object_predicate=self.object_predicate,
        )

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        subject, object = query_result_row

        row_output.add_triples.append(cast(Triple, (subject, self.connection, object)))
        row_output.instances_modified_count += 1
        return row_output



import json
import re
import sys
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable, Sequence, Set
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Generic, TypeVar

from cognite.client import CogniteClient
from cognite.client.data_classes._base import WriteableCogniteResource
from pydantic import AnyHttpUrl, ValidationError
from rdflib import RDF, XSD, Literal, Namespace, URIRef

from cognite.neat._constants import DEFAULT_NAMESPACE
from cognite.neat._graph.extractors._base import BaseExtractor
from cognite.neat._shared import Triple
from cognite.neat._utils.auxiliary import string_to_ideal_type

T_CogniteResource = TypeVar("T_CogniteResource", bound=WriteableCogniteResource)

DEFAULT_SKIP_METADATA_VALUES = frozenset({"nan", "null", "none", ""})

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


class InstanceIdPrefix(StrEnum):
    asset = "Asset_"
    label = "Label_"
    relationship = "Relationship_"
    sequence = "Sequence_"
    file = "File_"
    time_series = "TimeSeries_"
    event = "Event_"
    data_set = "DataSet_"

    @classmethod
    def from_str(cls, raw: str) -> "InstanceIdPrefix":
        raw = raw.title() + "_"
        if raw == "Timeseries_":
            return cls.time_series
        else:
            return cls(raw)


class ClassicCDFBaseExtractor(BaseExtractor, ABC, Generic[T_CogniteResource]):
    """This is the Base Extractor for all classic CDF resources.

    A classic resource is recognized in that it has a metadata attribute of type dict[str, str].

    Args:
        items (Iterable[T_CogniteResource]): An iterable of classic resource.
        namespace (Namespace, optional): The namespace to use. Defaults to DEFAULT_NAMESPACE.
        to_type (Callable[[T_CogniteResource], str | None], optional): A function to convert an item to a type.
            Defaults to None. If None or if the function returns None, the asset will be set to the default type.
        total (int, optional): The total number of items to load. If passed, you will get a progress bar if rich
            is installed. Defaults to None.
        limit (int, optional): The maximal number of items to load. Defaults to None. This is typically used for
            testing setup of the extractor. For example, if you are extracting 100 000 assets, you might want to
            limit the extraction to 1000 assets to test the setup.
        unpack_metadata (bool, optional): Whether to unpack metadata. Defaults to False, which yields the metadata as
            a JSON string.
        skip_metadata_values (set[str] | frozenset[str] | None, optional): If you are unpacking metadata, then
           values in this set will be skipped.
        camel_case (bool, optional): Whether to use camelCase instead of snake_case for property names.
            Defaults to True.
        as_write (bool, optional): Whether to use the write/request format of the items. Defaults to False.
    """

    _default_rdf_type: str
    _instance_id_prefix: str
    _SPACE_PATTERN = re.compile(r"\s+")

    def __init__(
        self,
        items: Iterable[T_CogniteResource],
        namespace: Namespace | None = None,
        to_type: Callable[[T_CogniteResource], str | None] | None = None,
        total: int | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
    ):
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.items = items
        self.to_type = to_type
        self.total = total
        self.limit = min(limit, total) if limit and total else limit
        self.unpack_metadata = unpack_metadata
        self.skip_metadata_values = skip_metadata_values
        self.camel_case = camel_case
        self.as_write = as_write

    def extract(self) -> Iterable[Triple]:
        """Extracts an asset with the given asset_id."""
        if self.total:
            try:
                from rich.progress import track
            except ModuleNotFoundError:
                to_iterate = self.items
            else:
                to_iterate = track(
                    self.items,
                    total=self.limit or self.total,
                    description=f"Extracting {type(self).__name__.removesuffix('Extractor')}",
                )
        else:
            to_iterate = self.items
        for no, asset in enumerate(to_iterate):
            yield from self._item2triples(asset)
            if self.limit and no >= self.limit:
                break

    def _item2triples(self, item: T_CogniteResource) -> list[Triple]:
        id_value: str | None
        if hasattr(item, "id"):
            id_value = str(item.id)
        else:
            id_value = self._fallback_id(item)
        if id_value is None:
            return []

        id_ = self.namespace[f"{self._instance_id_prefix}{id_value}"]

        type_ = self._get_rdf_type(item)

        # Set rdf type
        triples: list[Triple] = [(id_, RDF.type, self.namespace[type_])]
        if self.as_write:
            item = item.as_write()
        dumped = item.dump(self.camel_case)
        dumped.pop("id", None)
        # We have parentId so we don't need parentExternalId
        dumped.pop("parentExternalId", None)
        if "metadata" in dumped:
            triples.extend(self._metadata_to_triples(id_, dumped.pop("metadata")))

        triples.extend(self._item2triples_special_cases(id_, dumped))

        for key, value in dumped.items():
            if value is None or value == []:
                continue
            values = value if isinstance(value, Sequence) and not isinstance(value, str) else [value]
            for raw in values:
                triples.append((id_, self.namespace[key], self._as_object(raw, key)))
        return triples

    def _item2triples_special_cases(self, id_: URIRef, dumped: dict[str, Any]) -> list[Triple]:
        """This can be overridden to handle special cases for the item."""
        return []

    def _fallback_id(self, item: T_CogniteResource) -> str | None:
        raise AttributeError(
            f"Item of type {type(item)} does not have an id attribute. "
            f"Please implement the _fallback_id method in the extractor."
        )

    def _metadata_to_triples(self, id_: URIRef, metadata: dict[str, str]) -> Iterable[Triple]:
        if self.unpack_metadata:
            for key, value in metadata.items():
                if value and (self.skip_metadata_values is None or value.casefold() not in self.skip_metadata_values):
                    yield (
                        id_,
                        self.namespace[key],
                        Literal(string_to_ideal_type(value)),
                    )
        else:
            yield id_, self.namespace.metadata, Literal(json.dumps(metadata), datatype=XSD._NS["json"])

    def _get_rdf_type(self, item: T_CogniteResource) -> str:
        type_ = self._default_rdf_type
        if self.to_type:
            type_ = self.to_type(item) or type_
        return self._SPACE_PATTERN.sub("_", type_)

    def _as_object(self, raw: Any, key: str) -> Literal | URIRef:
        if key in {"data_set_id", "dataSetId"}:
            return self.namespace[f"{InstanceIdPrefix.data_set}{raw}"]
        elif key in {"assetId", "asset_id", "assetIds", "asset_ids", "parentId", "rootId", "parent_id", "root_id"}:
            return self.namespace[f"{InstanceIdPrefix.asset}{raw}"]
        elif key in {
            "startTime",
            "endTime",
            "createdTime",
            "lastUpdatedTime",
            "start_time",
            "end_time",
            "created_time",
            "last_updated_time",
        } and isinstance(raw, int):
            return Literal(datetime.fromtimestamp(raw / 1000, timezone.utc), datatype=XSD.dateTime)
        elif key == "labels":
            from ._labels import LabelsExtractor

            return self.namespace[f"{InstanceIdPrefix.label}{LabelsExtractor._label_id(raw)}"]
        elif key in {"sourceType", "targetType", "source_type", "target_type"} and isinstance(raw, str):
            # Relationship types. Titled so they can be looked up.
            return self.namespace[raw.title()]
        elif key in {"unit_external_id", "unitExternalId"}:
            try:
                return URIRef(str(AnyHttpUrl(raw)))
            except ValidationError:
                ...
        return Literal(raw)

    @classmethod
    def from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[T_CogniteResource], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
    ):
        total, items = cls._from_dataset(client, data_set_external_id)
        return cls(items, namespace, to_type, total, limit, unpack_metadata, skip_metadata_values, camel_case, as_write)

    @classmethod
    @abstractmethod
    def _from_dataset(
        cls, client: CogniteClient, data_set_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError

    @classmethod
    def from_hierarchy(
        cls,
        client: CogniteClient,
        root_asset_external_id: str,
        namespace: Namespace | None = None,
        to_type: Callable[[T_CogniteResource], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
    ):
        total, items = cls._from_hierarchy(client, root_asset_external_id)
        return cls(items, namespace, to_type, total, limit, unpack_metadata, skip_metadata_values, camel_case, as_write)

    @classmethod
    @abstractmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError

    @classmethod
    def from_file(
        cls,
        file_path: str | Path,
        namespace: Namespace | None = None,
        to_type: Callable[[T_CogniteResource], str | None] | None = None,
        limit: int | None = None,
        unpack_metadata: bool = True,
        skip_metadata_values: Set[str] | None = DEFAULT_SKIP_METADATA_VALUES,
        camel_case: bool = True,
        as_write: bool = False,
    ):
        total, items = cls._from_file(file_path)
        return cls(items, namespace, to_type, total, limit, unpack_metadata, skip_metadata_values, camel_case, as_write)

    @classmethod
    @abstractmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError
